
import os
import os.path as op
from shutil import which
import warnings

from monty.dev import requires
import paramiko
from collections import defaultdict

from pynter import SETTINGS, run_command
from pynter import LOCAL_DIR, REMOTE_DIR

def get_ssh_config(host,port=22,path=None):
    """
    Gets settings from config file for the given host.
    
    Parameters
    ----------
    host: (str) 
        The SSH host (can be an alias from ~/.ssh/config)
    port: (int)
        Default SSH port (default is 22).
    path: (str)
        Path of configuration file, if not provided uses the default ~/.ssh/config  
            
    Returns
    -------
    host: (str) 
        The SSH host.
    port: (int)
        SSH port.
    username: (str)
        SSH username (overrides config if provided)
    password: (str)
        SSH password.
    key_filename: (str)
        Path to private key file (overrides config if provided)
    """
    if path:
        path = os.path.abspath(path)
    ssh_config_path = path or os.path.expanduser("~/.ssh/config")

    if os.path.exists(ssh_config_path):
        ssh_config = paramiko.config.SSHConfig()
        with open(ssh_config_path) as f:
            ssh_config.parse(f)

        # Get the host-specific settings (if defined)
        host_config = ssh_config.lookup(host)

        host = host_config.get("hostname", host)
        port = int(host_config.get("port", port))
        username = host_config.get("user")
        key_filename = host_config.get("identityfile", [None])[0]

        return host, port, username, key_filename



class SSHProtocol:
    
    _instances = defaultdict(dict) #define attribute at class level, avoid having to define nested dict
    
    def __new__(cls, host, username=None, password=None, key_filename=None, port=22, timeout=10, reinitialize=False):
        """
        Creates or retrieves an SSHProtocol instance based on host and username.
        Ensures that when the same connection is not opened multiple times.
        If username is not provided, loads settings from  ~/.ssh/config.
        """
        if not username:
            host, port, username, key_filename = get_ssh_config(host)
        if username not in cls._instances[host] or reinitialize: 
            instance = super().__new__(cls)
            instance.host = host
            instance.username = username
            instance.password = password
            instance.key_filename = key_filename
            instance.port = port
            instance.timeout = timeout
            cls._instances[host][username] = instance
        return cls._instances[host][username]
    
    
    def __init__(self, 
                 host, 
                 username=None, 
                 password=None, 
                 key_filename=None, 
                 port=22, 
                 timeout=10, 
                 reinitialize=False):
        """
        Initializes an SSH connection, using settings from ~/.ssh/config if available.
        
        Parameters
        ----------

        host: (str) 
            The SSH host (can be an alias from ~/.ssh/config)
        username: (str)
            SSH username (overrides config if provided)
        password: (str)
            SSH password.
        key_filename: (str)
            Path to private key file (overrides config if provided)
        port: (int)
            SSH port (overrides config if provided)
        timeout: (int)
            Timeout for connection (default: 10 seconds)
        reinitialize: (bool)
            Force to create a new connection instead of using existing ones.
        """
        if not hasattr(self, 'initialized') or reinitialize:
            self.client = None
            self.initialized = True
            self.connect()
        else:
            if not self.client:
                print(f"Re-trying connection to {self.host}")
                self.connect()
                

    def __enter__(self):
        """Enter the context: return the instance itself."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the context: ensure the connection is closed."""
        self.close()


    def __repr__(self):
        lines = [f"{key}: {getattr(self,key)}" for key in ['host','port','username']]
        return '\n'.join(lines)
    
    def __print__(self):
        return self.__repr__()


    def close(self):
        """Closes the SSH connection."""
        if self.client:
            self.client.close()
            self.client = None


    def connect(self):
        """
        Establishes an SSH connection using the loaded settings, reuses if already connected.
        """
        if self.client and self.client.get_transport() and self.client.get_transport().is_active():
            print(f"Reusing existing connection to {self.host}")
            return 
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        try:
            self.client.connect(
                self.host,
                port=self.port,
                username=self.username,
                password=self.password,
                key_filename=self.key_filename,
                timeout=self.timeout
            )
        except Exception as e:
            print(f"Failed to connect: {e}")
            self.client = None


    def load_ssh_config(self):
        """
        Loads settings from ~/.ssh/config for the given host.
        """
        host, port, username, key_filename = get_ssh_config(self.host,self.port)
        self.host = host
        self.port = port
        self.username = username
        self.key_filename = key_filename


    def run_command(self,
                    command, 
                    timeout=10, 
                    printout=True, 
                    dry_run=False):
        """
        Executes a command over SSH with a timeout.
        """
        message = "%s@%s:" %(self.username,self.host),command
        if dry_run:
            if printout:
                print(message)
            return command, ''
        elif self.client:
            try:
                transport = self.client.get_transport()
                if transport and transport.is_active():
                    session = transport.open_session()
                    session.settimeout(timeout)
                    session.exec_command(command)
                    
                    stdout = session.makefile('r', -1).read().decode().strip() # decode string from b"" format
                    stderr = session.makefile_stderr('r', -1).read().decode().strip()
                    
                    if printout:
                        print(message)
                        if stdout:
                            print(stdout)
                        if stderr:
                            print("ERROR:",stderr)

                    return stdout, stderr

            except Exception as e:
                return None, f"Command execution failed: {e}"
        else:
            return None, "SSH client is not connected"


            


@requires(which("rsync"),
      "rsync needs to be installed, you can install it with 'sudo apt-get install rsync'.")

def rsync_from_hpc(
                    hostname=None,
                    remotedir=None,
                    localdir=None,
                    exclude=None,
                    printout=True,
                    dry_run=False,
                    makedir_on_local=True,
                    rsync_args={},
                    **kwargs
                    ):
    """
    Sync folders from HPC to local machine. The command "rsync" is used. With this function all
    the folders contained in the remote dir are synched to the local dir.

    Parameters
    ----------
    hostname : (str)
        Name of the host.
    remotedir : (str)
        Remote directory. The default is None. If None the one in config.yml file is used.
    localdir : (str)
        Local directory. The default is None. If None the one in config.yml file is used.
    exclude : (list)
        List of files to exclude in rsync. The default is None.
    dry_run : (bool)
        Perform dry run in rsync with --dry-run. The default is False. The dry_run in run_local is set to False.
    makedir_on_local : (bool)
        Make directory on local machine if it doesn't exist.
    rsync_args : (dict)
        Dictionary with arguments for rsync command. 
        If value is not none --key=value is added, else only --key is added.
    kwargs : (dict)
        Kwargs for the local run_command function.

    Returns
    -------
    stdout : (str)
        Output.
    stderr : (str)
        Error.
    """        
    cmd = get_rsync_from_hpc_command(hostname=hostname,
                                    remotedir=remotedir,
                                    localdir=localdir,
                                    exclude=exclude,
                                    dry_run=dry_run,
                                    makedir_on_local=makedir_on_local,
                                    rsync_args=rsync_args)   
    stdout,stderr = run_command(cmd,printout=printout,dry_run=False,**kwargs)
    return stdout,stderr


@requires(which("rsync"),
      "rsync needs to be installed, you can install it with 'sudo apt-get install rsync'.")

def rsync_to_hpc(
                hostname=None,
                localdir=None,
                remotedir=None,
                exclude=None,
                printout=True,
                dry_run=False,
                makedir_on_hpc=True,
                rsync_args={},
                **kwargs
                ):
    """
    Sync folders from local machine to HPC. The command "rsync" is used. With this function all
    the folders contained in the local dir are synched to the remote dir.

    Parameters
    ----------
    hostname : (str)
        Name of the host.
    localdir : (str)
        Local directory. The default is None. If None the one in config.yml file is used.
    remotedir : (str)
        Remote directory. The default is None. If None the one in config.yml file is used.
    exclude : (list)
        List of files to exclude in rsync. The default is None.
    dry_run : (bool)
        Perform dry run in rsync with --dry-run. The default is False. The dry_run in run_local is set to False.
    makedir_on_hpc : (bool)
        Make directory on HPC if it doesn't exist.
    rsync_args : (dict)
        Dictionary with arguments for rsync command. 
        If value is not none --key=value is added, else only --key is added.
    kwargs : (dict)
        Kwargs for the local run_command function.

    Returns
    -------
    stdout : (str)
        Output.
    stderr : (str)
        Error.
    """        
    cmd = get_rsync_to_hpc_command(hostname=hostname,
                                    localdir=localdir,
                                    remotedir=remotedir,
                                    exclude=exclude,
                                    dry_run=dry_run,
                                    makedir_on_hpc=makedir_on_hpc,
                                    rsync_args=rsync_args)    
    stdout,stderr = run_command(cmd,printout=printout,dry_run=False,**kwargs) #dry_run here is for the internal command, not rsync

    return stdout,stderr



def get_path_relative_to_hpc(path,localdir=None,remotedir=None):
    """
    Get path in HPC and relative to the local work directory.

    Parameters
    ----------
    path : (str)
        Path.
    localdir : (str)
        Path of work directory on local machine. If None the default from config file is used.
    remotedir : (str)
        Path of work directory on remote machine. If None the default from config file is used.

    Returns
    -------
    path_in_hpc : (str)
        Path in HPC.
    path_relative : (str)
        Path relative to work directory.
    is_path_on_hpc : (bool)
        Wether the given path's root is the HPC remote directory.
    """
    path = op.abspath(path)
    localdir = op.abspath(localdir) if localdir else LOCAL_DIR
    remotedir = op.abspath(remotedir) if remotedir else REMOTE_DIR
    if op.commonpath([remotedir,path]) == remotedir:
        is_path_on_hpc = True
    elif op.commonpath([localdir,path]) == localdir:
        is_path_on_hpc = False
    else:
        is_path_on_hpc = False
        warnings.warn('Path is not within local or remote work directory', UserWarning)
        
    if is_path_on_hpc:
        path_relative = op.abspath(path).replace(remotedir,'')
        path_in_hpc = path
    else:
        path_relative = op.abspath(path).replace(localdir,'')
        path_in_hpc = remotedir + path_relative
        
    return path_in_hpc, path_relative, is_path_on_hpc


def get_rsync_from_hpc_command(
                                hostname=None,
                                remotedir=None,
                                localdir=None,
                                exclude=None,
                                dry_run=False,
                                makedir_on_local=True,
                                rsync_args={}
                                ):
    """
    Parameters
    ----------
    hostname : (str)
        Name of the host.
    remotedir : (str)
        Remote directory. The default is None. If None the one in config.yml file is used.
    localdir : (str)
        Local directory. The default is None. If None the one in config.yml file is used.
    exclude : (list)
        List of files to exclude in rsync. The default is None.
    dry_run : (bool)
        Perform dry run in rsync with --dry-run. The default is False. The dry_run in run_local is set to False.
    makedir_on_local : (bool)
        Make directory on local machine if it doesn't exist.
    rsync_args : (dict)
        Dictionary with arguments for rsync command. 
        If value is not none --key=value is added, else only --key is added.
    """
    hostname = hostname if hostname else SETTINGS['HPC']['hostname']
    localdir = localdir or os.getcwd()
    remotedir = remotedir or get_path_relative_to_hpc(path=localdir)[0]
    localdir = op.abspath(localdir)
    remotedir = op.join(remotedir,'')  #ensure backslash at the end
    localdir = op.join(localdir,'')
    
    cmd = ''
    if makedir_on_local:
        cmd += 'mkdir -p %s' %localdir
        cmd += ' ; '
    
    cmd += "rsync -r -uavzh " #keep the spaces
    if dry_run:
        cmd += "--dry-run "
        
    if exclude:
        for s in exclude:
            cmd += f'--exclude={s} ' 
    
    for key,value in rsync_args.items():
        if value is not None:
            cmd += f'--{key}={value} ' #keep the spaces at the end
        else:
            cmd += f'--{key} '
    
    cmd += f"-e ssh {hostname}:{remotedir} {localdir} "
    return cmd


def get_rsync_to_hpc_command(
                                hostname=None,
                                localdir=None,
                                remotedir=None,
                                exclude=None,
                                dry_run=False,
                                makedir_on_hpc=True,
                                rsync_args={}
                                ):
    """
    Parameters
    ----------
    hostname : (str)
        Name of the host.
    localdir : (str)
        Local directory. The default is None. If None the one in config.yml file is used.
    remotedir : (str)
        Remote directory. The default is None. If None the one in config.yml file is used.
    exclude : (list)
        List of files to exclude in rsync. The default is None.
    dry_run : (bool)
        Perform dry run in rsync with --dry-run. The default is False. The dry_run in run_local is set to False.
    makedir_on_hpc : (bool)
        Make directory on HPC if it doesn't exist.
    rsync_args : (dict)
        Dictionary with arguments for rsync command. 
        If value is not none --key=value is added, else only --key is added.
    """
    hostname = hostname or SETTINGS['HPC']['hostname']
    localdir = localdir or os.getcwd()
    remotedir = remotedir or get_path_relative_to_hpc(path=localdir)[0]
    localdir = op.abspath(localdir)
    localdir = op.join(localdir,'')  #ensure backslash at the end
    remotedir = op.join(remotedir,'')
    
    if makedir_on_hpc:
        mkdir_command = 'mkdir -p %s' %remotedir
        if dry_run:
            print(f'DRY RUN: {mkdir_command}')
        else:
            ssh = SSHProtocol(hostname)
            ssh.run_command(mkdir_command)
        
    
    cmd = "rsync -r -uavzh " #keep the spaces
    if dry_run:
        cmd += "--dry-run "
        
    if exclude:
        for s in exclude:
            cmd += f'--exclude={s} '
            
    for key,value in rsync_args.items():
        if value is not None:
            cmd += f'--{key}={value} ' #keep the spaces at the end
        else:
            cmd += f'--{key} '
            
    cmd += f"-e ssh {localdir} {hostname}:{remotedir} "
    return cmd






from abc import abstractmethod
import os
import os.path as op
import shutil
import warnings

from pynter import SETTINGS, LOCAL_DIR, REMOTE_DIR
from pynter.tools.utils import get_object_feature
from pynter.hpc.slurm import JobSettings
from pynter.hpc.ssh import rsync_from_hpc, rsync_to_hpc, get_path_relative_to_hpc


from pynter.jobs.manager import JobManager


class Job:
    
    def __init__(self,
                 path=None,
                 inputs=None,
                 job_settings=None,
                 outputs=None,
                 job_script_filename=None,
                 name=None,
                 hostname=None,
                 localdir=None,
                 remotedir=None):
        """
        Class to control and organize inputs and outputs of a generic job.

        Parameters
        ----------
        path : (str), optional
            Path where job is stored. The default is None. If None the work dir is used.
        inputs : (dict), optional
            Dictionary with input data. The default is None.
        job_settings : (dict), optional
            Dictionary with job settings. The default is None. Documentation in JobSettings class in hpc.slurm module
        outputs : (dict), optional
            Dictionary with output data. The default is None.
        job_script_filename : (str), optional
            Filename of job script. The default is taken from the key 'filename' in the job_settings in the config file.
        name : (str)
            Name of the job. If none the name is searched in the job script.
        """
                
        self.path = op.abspath(path) if path else os.getcwd()
        self.inputs = inputs
        self.job_settings = JobSettings(**job_settings) if job_settings else JobSettings() 
        self.outputs = outputs
        self.job_script_filename = job_script_filename if job_script_filename else SETTINGS['job_script_filename']
        
        self.hostname = hostname or SETTINGS['HPC']['hostname']
        self.localdir = localdir or LOCAL_DIR 
        self.remotedir = remotedir or REMOTE_DIR
        
        self.path_in_hpc, self.path_relative, self.is_path_on_hpc = get_path_relative_to_hpc(
                                                        path=self.path,
                                                        localdir=self.localdir,
                                                        remotedir=self.remotedir)

        if outputs:
            self.get_output_properties()
         
        if name:
            self.name = name
        elif self.job_settings:
            self.name = self.job_settings['sbatch']['job-name']
        elif op.isfile(op.join(self.path,self.job_script_filename)):
            s = JobSettings.from_bash_file(self.path,filename=self.job_script_filename)
            self.name = s.settings['sbatch']['job-name']
        else:
            self.name = 'no_name'
            
        if not self.job_settings:
            self.job_settings = {}
        self.job_settings['sbatch']['job-name'] = self.name


    def __str__(self):
        jobclass = self.jobclass
        if hasattr(self,'group'):
            if self.group != '':
                printout = '%s "%s" of group "%s"' %(jobclass, self.name, self.group)
            else:
                printout = '%s "%s"' %(jobclass, self.name)
        else:
            self.group = ''
            printout = '%s "%s"' %(jobclass, self.name)
        
        return printout
    
    def __repr__(self):
        return self.__str__()
        
    @property
    def jobclass(self):
        return self.__class__.__name__

    @property
    def job_id(self):
        if not self._job_id:
            job_manager = JobManager(self)
            stdout,stderr = job_manager.qstat()
            job_id = job_manager.get_job_id_from_queue(stdout)
            self._job_id = job_id
        return self._job_id
        
    
    def add_to_asedb(self,db,
                     duplicate=False,
                     converged_only=True,
                     verbose=True,
                     properties_to_write=[]):
        """
        Add Atoms object in self.outputs to ASE database

        Parameters
        ----------
        db : (ase database)
            ASE database object (SQLite3Database).
        duplicate : (bool)
            Add job to db even if an entry with same path is already present.
        converged_only : (bool)
            Add job only if job.is_converged is True.
        verbose : (bool)
            Print info on screen.
        properties_to_write: (list)
            List of job properties to add to ASE DB.
            kwargs of the form feature=get_object_feature(job,feature) will
            be added to ase.db.write. See docs of get_object_feature for details.
        """
        printout = []
        properties_to_write.append('path')
        if not converged_only or self.is_converged:
            if 'Atoms' in self.outputs.keys():
                if duplicate or all(self.path != row.path for row in db.select()):
                    try:
                        atoms = self.outputs['Atoms']
                        kwargs={}
                        for feature in properties_to_write:
                            if isinstance(feature,list):
                                key = feature[0]
                                for k in feature[1:]:
                                    key = key + '_%s'%k
                            else:
                                key = feature
                            kwargs[key] = get_object_feature(self,feature)
                        db.write(atoms=atoms,**kwargs)
                        printout.append(f'Job added to DB {self.path}')
                    except Exception as e:
                        print(e)
                        printout.append(f'Adding Atoms object to DB failed for {self}')
                else:
                    printout.append(f'Skipping job already in DB for {self.path}')
            else:
                printout.append(f'Atoms object not in outputs dict for {self.path}')
        else:
            printout.append(f'Skipping unconverged self at {self.path}')
        
        if verbose:
            for line in printout:
                print(line)
    
    
    def cancel_job(self):
        """
        Cancel job on HPC
        """      
        JobManager(self).scancel()    
        return 


    def copy(self):
        """
        Copy Job object
        """
        path=self.path
        inputs=self.inputs.copy() if self.inputs else None
        job_settings=self.job_settings.copy() if self.job_settings else None
        outputs=self.outputs.copy() if self.outputs else None
        job_script_filename=self.job_script_filename
        name=self.name
        return self.__class__(path=path,inputs=inputs,job_settings=job_settings,
                       outputs=outputs,job_script_filename=job_script_filename,
                       name=name)
    
    
    def delete_job_files(self,safety=True):
        """
        Delete Job folder (self.path)

        Parameters
        ----------
        safety : (bool), optional
            Ask confirmation to delete job. The default is True.
        """
        if safety:
            inp = input('Are you sure you want to delete Job %s ? (y/n) : ' %self.name)
            if inp in ('y','Y'):
                shutil.rmtree(self.path)
                print('Deleted Job %s'%self.name)
            else:
                print('Job %s is left unchanged'%self.name)
        else:
            shutil.rmtree(self.path)
            print('Deleted Job %s'%self.name)
        return


    @abstractmethod
    def get_inputs(self):
        pass

    @abstractmethod
    def get_outputs(self):
        pass
    
    @abstractmethod
    def get_output_properties(self):
        pass

    @abstractmethod
    def insert_in_database(self):
        pass


    def job_queue(self):
        """
        Print job queue from HPC on screen
        
        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.
        """
        stdout,stderr = JobManager(self).qstat()        
        return stdout,stderr
        

    def run_job(self,write_input=True,sync=True):
        """
        Run job on HPC. Input files are automatically written and sync to HPC is performed.

        Parameters
        ----------
        write_input : (bool), optional
            Write input file stored in "inputs" dictionary. The default is True.
        sync : (bool), optional
            Sync files to HPC before running. The default is True

        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.
        """
        if write_input:
            self.write_input()
        if sync and not self.is_path_on_hpc:
            self.sync_to_hpc()
        stdout, stderr = JobManager(self).sbatch()       
        return stdout,stderr
    

    def sync_from_hpc(self,stdouts=False,exclude=None,dry_run=False):
        """
        Sync job data from HPC to local machine

        Parameters
        ----------
        stdouts : (bool), optional
            Return output and error strings. The default is False.
        exclude : (list), optional
            List of files to exclude in rsync. The default is None.
        dry_run : (bool), optional
            Perform dry run in rsync with --dry-run. The default is False.

        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.

        """
        if self.is_path_on_hpc:
            raise ValueError('Cannot sync from remote machine, only from local')
        localdir = op.abspath(self.path)
        stdout,stderr = rsync_from_hpc(hostname=self.hostname,
                                       remotedir=self.path_in_hpc,
                                       localdir=localdir,
                                       exclude=exclude,
                                       dry_run=dry_run)
        if stdouts:
            return stdout,stderr
        else:
            return
        
        
    def sync_to_hpc(self,stdouts=False,exclude=None,dry_run=False):
        """
        Sync job data from local machine to HPC

        Parameters
        ----------
        stdouts : (bool), optional
            Return output and error strings. The default is False.
        exclude : (list), optional
            List of files to exclude in rsync. The default is None.
        dry_run : (bool), optional
            Perform dry run in rsync with --dry-run. The default is False.

        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.

        """
        if self.is_path_on_hpc:
            raise ValueError('Cannot sync from remote machine, only from local')
        localdir = op.abspath(self.path)
        stdout,stderr = rsync_to_hpc(hostname=self.hostname,
                                     localdir=localdir,
                                     remotedir=self.path_in_hpc,
                                     exclude=exclude,
                                     dry_run=dry_run)
        if stdouts:
            return stdout,stderr
        else:
            return


    def status(self):
        """
        Get job status from HPC. Gets queue from HPC and gets Job status from the queue.

        Returns
        -------
        status : (str)
            Job status. Possible status are 'PENDING','RUNNING','NOT IN QUEUE'.
        """
        job_manager = JobManager(self)
        stdout,stderr = job_manager.qstat()
        status = job_manager.get_status_from_queue()            
        return status
 
           
    @abstractmethod
    def write_input():
        pass
        


def get_job_from_directory(path=None,
                           job_script_filename=None,
                           load_outputs=True,
                           hostname=None,
                           localdir=None,
                           remotedir=None,
                           **kwargs):   
    """
    Get Job object from a directory. The job type is selected based on the content of the folder.
    - If ("INCAR","KPOINTS","POSCAR","POTCAR") files are present, VaspJob is initialized.
    - If ("INCAR","KPOINTS","POTCAR") files are present and "POSCAR" is not present, VaspNEBJob is initialized.
    - If ("input.in") files are present, LammpsJob is initialized. In case input, data and log files are named
      differently than "input.in", "structure.data" and "log.lammps", set it in jobs_kwargs as:
          jobs_kwargs = {'LammpsJob':{
                                      'input_filename':<input filename>,
                                      'data_filename':<data filename>,
                                      'log_filename':<log filename>
                                      }

    Parameters
    ----------
    path : (str)
        Directory containing job.
    job_script_filename : (str), optional
        Filename of job script. The default is taken from the key 'filename' in the job_settings in the config file.
    sort : (str or list), optional
        Sort list of jobs by features. If False or None jobs are not sorted. The default is 'name'.
    load_outputs : (bool)
        Load job outputs. The default is True.
    kwargs : (dict)
        Kwargs to be used when importing job from directory for each job class.
        If only one job class is imported use the kwargs as usual.
        If more than 1 job class is imported use it as:
            JobClass1=jobclass1_kwargs, JobClass2=jobclass2_kwargs

    Returns
    -------
    jobs : (list)
        List of Job objects.
    """    
    from pynter.jobs.vasp.vasp_jobs import VaspJob, VaspNEBJob
    from pynter.jobs.lammps.lammps_jobs import LammpsJob, input_default_filename
    
    path = path if path else os.getcwd()
    job_script_filename = job_script_filename if job_script_filename else SETTINGS['job_script_filename']
    
    if kwargs:
        if all([key not in ['VaspJob','VaspNEBJob','LammpsJob'] for key in kwargs.keys()]):
            warnings.warn('Job class not specified in kwargs, only do this if you are not looping over different job classes')
    
    for root,dirs,files in os.walk(path):   
        # VASP
        if all(f in files for f in ['INCAR','KPOINTS','POSCAR','POTCAR']):
            path = op.abspath(root)
            if kwargs:
                vaspjob_kwargs = kwargs['VaspJob'] if 'VaspJob' in kwargs.keys() else kwargs
            else:
                vaspjob_kwargs = {}
            job = VaspJob.from_directory(
                        path=path,
                        job_script_filename=job_script_filename,
                        load_outputs=load_outputs,
                        hostname=hostname,
                        localdir=localdir,
                        remotedir=remotedir,
                        **vaspjob_kwargs)
            return job

        # VASP + NEB
        elif all(f in files for f in ['INCAR','KPOINTS','POTCAR']) and 'POSCAR' not in files:
            path = op.abspath(root)
            job = VaspNEBJob.from_directory(
                                        path=path,
                                        job_script_filename=job_script_filename,
                                        load_outputs=load_outputs,
                                        hostname=hostname,
                                        localdir=localdir,
                                        remotedir=remotedir)
            return job

        # LAMMPS
        else:
            if kwargs:
                lammpsjob_kwargs = kwargs['LammpsJob'] if 'LammpsJob' in kwargs.keys() else kwargs
            else:
                lammpsjob_kwargs = {}
            lammps_input_filename = lammpsjob_kwargs['input_filename'] if 'input_filename' in lammpsjob_kwargs.keys() else input_default_filename
            if lammps_input_filename in files:
                job = LammpsJob.from_directory(
                                            path=path,
                                            job_script_filename=job_script_filename,
                                            load_outputs=load_outputs,
                                            hostname=hostname,
                                            localdir=localdir,
                                            remotedir=remotedir,
                                            **lammpsjob_kwargs)
                return job


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 30 15:30:49 2023

@author: villa
"""
import os

from pynter import SETTINGS
from pynter.jobs.vasp.vasp_automations import VaspAutomation
from pynter.jobs.datasets import Dataset

def parse_common_args(parser):
    auto = VaspAutomation()
    
    parser.add_argument('-j','--job-script',help='Job script filename (default: %(default)s)',required=False,default=SETTINGS['job_script_filename'],type=str,metavar='',dest='job_script_filename')
    parser.add_argument('-s','--status',help='Write exit status to file (default: %(default)s)',required=False,default=auto.status_filename,type=str,metavar='',dest='status_filename')
    parser.add_argument('-e','--error-check',action='store_true',help='Perform error checking (default: %(default)s)',required=False,default=False,dest='error_check')
    return parser


def setup_automation(subparsers):
    
    subparsers_automation = subparsers.add_parser('automation',help='Set up automation for calculations')
    subparsers_automation = subparsers_automation.add_subparsers()
    
    parser_vasp = subparsers_automation.add_parser('vasp',help='VASP automation')
    parser_vasp = parse_common_args(parser_vasp)
    parser_vasp.add_argument('-c','--contcar',action='store_true',help='Copy CONTCAR to POSCAR of next step',required=False,default=False,dest='contcar')
    parser_vasp.add_argument('-W','--wavecar',action='store_true',help='Copy WAVECAR to next step',required=False,default=False,dest='wavecar')
    parser_vasp.add_argument('-C','--chgcar',action='store_true',help='Copy CHGCAR to next step',required=False,default=False,dest='chgcar')
    parser_vasp.add_argument('-K','--check-kpoints',action='store_true',help='Copy WAVECAR and POSCAR only if KPOINTS of next step are the same',required=False,default=False,dest='check_kpoints')
    parser_vasp.set_defaults(func=run_automation_vasp)
    
    parser_vasp_neb = subparsers_automation.add_parser('vasp-NEB',help='VASP automation for NEB calculations')
    parser_vasp_neb = parse_common_args(parser_vasp_neb)
    parser_vasp_neb.add_argument('-c','--contcar',action='store_true',help='Copy CONTCAR to POSCAR of next step',required=False,default=False,dest='contcar')
    parser_vasp_neb.add_argument('-W','--wavecar',action='store_true',help='Copy WAVECAR to next step',required=False,default=False,dest='wavecar')
    parser_vasp_neb.add_argument('-C','--chgcar',action='store_true',help='Copy CHGCAR to next step',required=False,default=False,dest='chgcar')
    parser_vasp_neb.add_argument('-K','--check-kpoints',action='store_true',help='Copy WAVECAR and POSCAR only if KPOINTS of next step are the same',required=False,default=False,dest='check_kpoints')
    parser_vasp_neb.set_defaults(func=run_automation_vasp_neb)


def run_automation_vasp(args):
    automation = VaspAutomation()
    ds = Dataset.from_directory(
                                path='../',
                                job_script_filenames=args.job_script_filename,
                                sort='path')
    path = os.getcwd()
    job_current = ds.select_jobs(path=path)[0]
    is_converged_electronic, is_converged_ionic = automation.check_convergence(job_current)
    if is_converged_electronic and is_converged_ionic:
        current_index = ds.jobs.index(job_current)
        job_next = ds.jobs[current_index + 1]
        files = []
        if args.contcar:
            files.append('CONTCAR')
        if args.chgcar:
            files.append('CHGCAR')
        if args.wavecar:
            files.append('WAVECAR')
        automation.copy_files_from_job1_to_job2(
                                    job1=job_current,
                                    job2=job_next,
                                    filenames=files,
                                    check_kpoints=args.check_kpoints)
        
        job_next.run_job(sync=False,write_input=False)
    automation.write_status(path=path)


def run_automation_vasp_neb(args):
    raise NotImplementedError('VASP NEB automation still to be implemented')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 10 13:43:48 2025

@author: villa
"""
import numpy as np

from pymatgen.io.vasp.inputs import Kpoints

from pynter.vasp.schemes.core import InputSets


class VariationsSchemes(InputSets):


    def charge_states(self,charges,locpot=True):
        """
        Generate calculation schemes for calculations of different charge states.
        Only the NELECT parameter in INCAR is varied.
        A list of desired charges (float or int) is required.
        
        Returns:
            List of VaspJob objects        
        """  
        jobs = []        
        val = {}
        for p in self.potcar:
            val[p.element] = p.nelectrons        
        nelect = sum([ val[el.symbol]*coeff for el,coeff in self.structure.composition.items()])
        
        for q in charges:
            if isinstance(q,float) and not q.is_integer():
                q = np.around(q,decimals=1)
            stepname = f'q{q}'
            vaspjob = self.get_vaspjob(add_to_job_name=stepname,add_to_job_path=stepname)
            vaspjob.incar['NELECT'] = nelect - q 
            if locpot:
                vaspjob.incar['LVTOT'] = '.TRUE.' #most likely needed for corrections in defect calculations
            jobs.append(vaspjob)
            
        return jobs    


    def convergence_encut(self,encuts=[]):
        """
        Scheme for energy cutoff convergence

        Parameters
        ----------
        encuts : (list), optional
            List of energy cutoffs to use. If not provided a range from 200 to 700 eV is used.The default is [].

        Returns
        -------
        jobs : (list)
            List of VaspJob objects.
        """
        jobs = []        
        if not encuts:
            encuts = range(300,800,100)

        for ec in encuts:           
            print(ec)
            stepname = f'cutoff{ec}'
            vaspjob = self.pbe_scf(add_to_job_name=stepname,add_to_job_path=stepname)            
            vaspjob.incar['ENCUT'] = ec
            jobs.append(vaspjob)
            
        return jobs
        

    def convergence_kpoints(self,kpoints_meshes=[]):
        """
        Scheme for kpoints convergence.

        Parameters
        ----------
        kpoints_meshes : (list), optional
            List of kpoints meshes to use. If not provided a range from 2x2x2 to 7x7x7 is used.The default is [].

        Returns
        -------
        jobs : (list)
            List of VaspJob objects.
        """
        jobs = []        
        if not kpoints_meshes:
            kpoints_meshes = []
            for k in range(2,9):
                kpoints_meshes.append((k,k,k))

        for kmesh in kpoints_meshes:
            stepname = 'k%ix%ix%i' %(kmesh[0],kmesh[1],kmesh[2])
            vaspjob = self.get_vaspjob(add_to_job_name=stepname,add_to_job_path=stepname)
            vaspjob.incar.pop('KPAR', None)
            vaspjob.incar['NSW'] = 0
            vaspjob.inputs['KPOINTS'] = Kpoints().gamma_automatic(kpts=kmesh)
            jobs.append(vaspjob)
            
        return jobs
            


    def fractional_charge_linearity(self,nelect=None):
        """
        Generate calculation scheme for occupation linearity test.
        The number of electrons are scanned from NELECT to NELECT + 1 with interval of 0.2.
        
        Parameters
        --------
        nelect: (int)
            NELECT to start from. If None the total valence electrons from Potcar are used.
        Returns:
            List of VaspJob objects
        """        
        jobs = []        
        val = {}
        for p in self.potcar:
            val[p.element] = p.nelectrons 
        if not nelect:
            nelect = sum([ val[el]*self.structure.composition.as_dict()[el] for el in self.structure.composition.as_dict()])
        
        for q in np.arange(0,1.2,0.2):
            q = np.around(q,decimals=1)
            stepname = f'q{q}'
            vaspjob = self.get_vaspjob(add_to_job_name=stepname,add_to_job_path=stepname)
            vaspjob.incar['NELECT'] = nelect + q
            jobs.append(vaspjob)
            
        return jobs



import os.path as op
from pynter import run_command
from pynter.hpc.ssh import SSHProtocol
from pynter import SETTINGS
from pynter.tools.utils import grep_list
         

def get_mkdir_command(path):
    """
    Get command to make a new directory if doesn't exist.

    Parameters
    ----------
    path : (str)

    Returns
    -------
    cmd : (str)
        Command.
    """
    path = op.abspath(path)
    cmd = 'mkdir -p %s' %path
    return cmd

def get_qstat_command():
    """
    Get command to view job queue.

    Returns
    -------
    cmd : (str)
        Command.
    """
    cmd = 'squeue -o "%.10i %.9P %.40j %.8u %.2t %.10M %.5D %R"'
    return cmd

def get_sbatch_command(path,job_script_filename='job.sh'):
    """
    Execute "sbatch" command to run job.

    Parameters
    ----------
    path : (str), optional
        Path where the job script is stored (relative or absolute).
    job_script_filename : (str), optional
        Filename of the job script. The default is 'job.sh'.

    Returns
    -------
    cmd (str)
        Command.
    """   
    path = op.abspath(path)
    cmd = f'cd {path} ; sbatch {job_script_filename}'
    return cmd
            
def get_scancel_command(*args):
    """
    Get command to cancel jobs using scancel

    Parameters
    ----------
    *args : (list)
        Job-IDs.
    
    Returns
    -------
    cmd : (str)
        Command.
    """
    cmd = 'scancel '
    for arg in args:
        cmd += arg + ' '
    return cmd   



class JobManager:
    
    
    def __init__(self,job,hostname=None):
        """
        Class to run and manage job both remotely and/or locally.
        """
        self.hostname = hostname or job.hostname
        self.job = job


    def _run_job_command(self,cmd,printout=True,dry_run=False,**kwargs):
        """
        Execute command related to Job. If the current location is on HPC cluster
        the command is run on the local OS, otherwise is run on the remote HPC with ssh.
        """
        if self.job.is_path_on_hpc:
            stdout,stderr = run_command(cmd,printout=printout,dry_run=dry_run)   
        else:
            ssh = SSHProtocol(self.hostname)
            stdout, stderr = ssh.run_command(cmd,printout=printout,dry_run=dry_run)
        return stdout, stderr


    def get_job_id_from_queue(self,stdout):
        """
        Get job ID from the queue on HPC. qstat is not explicitly called here to be able to handle
        looping without running qstat every time.
        """        
        queue = stdout.splitlines()
        job_lines = grep_list(self.job.name,queue)
        if job_lines == []:
            raise ValueError (f'Job named "{self.name}" is not currently running or pending')
        elif len(job_lines) > 1:
            raise ValueError (f'More than one job named "{self.name}" has been found in queue:\n{stdout}')
        else:
            job_line = job_lines[0].split()
            job_id = job_line[0]
        
        return job_id


    def get_status_from_queue(self,stdout):     
        """
        Get Job status from stdout of qstat command in HPC. 
        qstat is not explicitly called here to be able to handle
        looping without running qstat every time.
    
        Returns
        -------
        status : (str)
            Job status. Possible status are 'PENDING','RUNNING','NOT IN QUEUE'.
        """
        queue = stdout.splitlines()
        job_lines = grep_list(self.job.name,queue)
        if job_lines == []:
            status = 'NOT IN QUEUE'
        elif len(job_lines) > 1:
            raise ValueError (f'More than one job named "{self.name}" has been found in queue:\n{stdout}')
        else:
            job_line = job_lines[0].split()
            status = job_line[4]
            if status == 'PD':
                status = 'PENDING'
            if status == 'R':
                status = 'RUNNING'
            if status == 'CG':
                status = 'COMPLETED'
            
        return status
    

    def mkdir(self,printout=True,dry_run=False):
        """
        Make new directory if doesn't exist.

        Parameters
        ----------
        path : (str)
            Path (relative or absolute.
        printout : (bool), optional
            Write output on screen. The default is True.
        dry_run : (bool), optional
            Only return the command to be run and empty error string. The default is False.

        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.
        """
        cmd = get_mkdir_command(self.job.path)
        stdout, stderr = self._run_job_command(cmd,printout=printout,dry_run=dry_run)
        return stdout, stderr
        
        
    def qstat(self,printout=True,dry_run=False):
        """
        Check queue status.

        Parameters
        ----------
        cmd : (str), optional
            Command to run. The default is 'squeue -o "%.10i %.9P %.40j %.8u %.2t %.10M %.5D %R"'.
        printout : (bool), optional
            Write output on screen. The default is True.
        dry_run : (bool), optional
            Only return the command to be run and empty error string. The default is False.

        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.
        """        
        cmd = get_qstat_command()
        stdout, stderr = self._run_job_command(cmd,printout=printout,dry_run=dry_run)
        return stdout, stderr


    def sbatch(self,printout=True,dry_run=False):
        """
        Execute "sbatch" command to run job.

        Parameters
        ----------
        printout : (bool), optional
            Write output on screen. The default is True.
        dry_run : (bool), optional
            Only return the command to be run and empty error string. The default is False.

        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.
        """   
        if self.job.is_path_on_hpc:
            path = self.job.path
        else:
            path = self.job.path_in_hpc
        cmd = get_sbatch_command(path=path,job_script_filename=self.job.job_script_filename)
        stdout, stderr = self._run_job_command(cmd,printout=printout,dry_run=dry_run)     # I've used the run local because there was a probelm with the multiple command given with ;
        return stdout,stderr                                            # to check again if possible
        
        
    def scancel(self,printout=True,dry_run=False):
        """
        Cancel jobs using scancel

        Parameters
        ----------
        *args : (str)
            Job-id.
        printout : (bool), optional
            Write output on screen. The default is True.
        dry_run : (bool), optional
            Only return the command to be run and empty error string. The default is False.
        """
        job_id = self.job.job_id
        cmd = get_scancel_command(job_id)
        stdout, stderr = self._run_job_command(cmd,printout=printout,dry_run=dry_run)
        return stdout,stderr        
        
    
    